// elevator.h 
//	Data structures to emulate a bank of elevators.
//
//	There are two interfaces -- one for the threads controlling
//	the elevators, and one for threads representing elevator
//	passengers.  The only communication between these two
//	groups is via the elevator device.  The elevator device
//	interface has both regular calls from the rider/controller
//	threads to the device, as well as "callbacks" (interrupts) 
//	from the device to the riders/controllers when events occur.
//
//  DO NOT CHANGE -- part of the machine emulation
//
// Copyright (c) 1992-1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#ifndef ELEVATORBANK_H
#define ELEVATORBANK_H

#include "copyright.h"
#include "utility.h"
#include "callback.h"
#include "list.h"
#include "elevatortypes.h"
#include "elevatorinterface.h"
#include "riderinterface.h"

class ElevatorInfo;
class PendingElevatorEvent;
class ElevatorBank;
class StandardElevatorInterface;
class StandardRiderInterface;

typedef List<PendingElevatorEvent *> ListOfEvents;

// The following class defines a bank of elevators
class ElevatorBank : public CallBackObj {
public:
  // Initialize the elevator hardware, specfying # of elevators and # of floors
  ElevatorBank(int numElvtrs, int numFlrs);

  // deallocate the elevator hardware  
  ~ElevatorBank();

  // return an object that can be used to communicate with the elevator HW
  ElevatorInterface* getElevatorInterface() {
    return elevIF; }

  // return an object that can be used to communicate with the rider "HW"
  RiderInterface* getRiderInterface() {
    return riderIF; }

private:  
  // Controller interface to elevator hardware; access these methods through
  // an ElevatorInterface object

  void OpenDoors(int elevator);
  void CloseDoors(int elevator);
  void MoveTo(int goingToFloor, int elevator);
  int  WhereIsElevator(int elevator);
  void MarkDirection(int elevator, Direction dir);

  // Rider interface to elevator control panel
 
  void PressButton(int onFloor, Direction goingTo);
  Direction getDirection(int elevator);
  bool EnterElevator(int onFloor, int elevator);
  void PressFloor(int goingToFloor, int onElevator);
  bool ExitElevator(int onFloor, int elevator);

  int numElevators;		// how many elevators in this bank?
  int numFloors;		// how many floors in this building?

  ElevatorInterface* elevIF;	// provides access to elevator methods
  RiderInterface* riderIF;	// provides access to rider methods

  ListOfEvents *riderEvents;	// pending events relevant to riders
  ListOfEvents *controllerEvents;// pending events relevant to controllers
  ElevatorInfo **elevators;	// array of per-elevator state

  void CallBack();		// called internally when the elevator
    				// hardware generates an interrupt
    
  void PostEvent(ListOfEvents *list, ElevatorEvent event, int floor,
		 int elevator, bool inHandler);	
    				// an event occurred that requires
				// a callback
				
  ElevatorEvent getNextEvent(ListOfEvents *list, int *floor, int *elevator);
				// pull next event off the pending list
				// of events, return NoEvent if list is
				// empty

  friend class StandardElevatorInterface;
  friend class StandardRiderInterface;
};

#endif // ELEVATORBANK_H
