// elevatortest.cc 
//	Driver code to test the elevator device implementation.
//	We create a single rider thread and a single elevator thread;
//	in practice there will be multiple riders and multiple elevators,
//	each with its own thread.
//
// Copyright (c) 1996 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"
#include "elevatortest.h"
#include "elevatorbank.h"
#include "synch.h"
#include "thread.h"

// Data structures to control elevator device self test.
// Definition private to this module. 
class ElevatorInspector : public CallBackObj{
public:
  ElevatorInspector();		// allocate self test data structures
  ~ElevatorInspector();		// deallocate self test data
    
  void RiderTest();		// test out the elevator simulation,
  				// by acting like a rider
  void ControllerTest();	// test out elevator simulation, by
    				// acting like an elevator controller
  void CallBack();		// get notification of event
    
private:
  ElevatorBank* elevators;
  ElevatorInterface* elevatorInterface;
  RiderInterface* riderInterface;
    
  Semaphore *riderWakeup;	// to synchronize rider with callbacks
  Semaphore *controllerWakeup;	// to synch controller with callbacks
    
  // The next two functions wait for the rider/controller
  // to be signalled.  In a real implementation there would be
  // multiple riders and multiple elevators, so some way of getting
  // each event to the right set of threads is needed.
  ElevatorEvent WaitForNextRiderEvent(int *floor, int *elevator) {
    while (1) {
      ElevatorEvent event = riderInterface->getNextEvent(floor, elevator);

      if (event != NoEvent)
	return event;
      else
	riderWakeup->P();
    }
  }

  ElevatorEvent WaitForNextControllerEvent(int *floor, int *elevator) {
    while (1) {
      ElevatorEvent event = elevatorInterface->getNextEvent(floor, elevator);

      if (event != NoEvent)
	return event;
      else
	controllerWakeup->P();
    }
  }

};

static void
ControllerTest(ElevatorInspector *inspector) {
  inspector->ControllerTest();
}

//----------------------------------------------------------------------
// ElevatorSelfTest
//	Run self test on elevator device.  Create one thread to
//	to act as the rider, and another to act as the controller.
//----------------------------------------------------------------------

void
ElevatorSelfTest () {
  ElevatorInspector *inspector = new ElevatorInspector();
  Thread *controller = new Thread("controller");
    
  controller->Fork((void (*)(void*)) ControllerTest, (void*) inspector);
  inspector->RiderTest();
  delete inspector;
}

//----------------------------------------------------------------------
// ElevatorInspector::ElevatorInspector
//	Allocate and initialize data structures for testing the
//	elevator simulation.
//----------------------------------------------------------------------

ElevatorInspector::ElevatorInspector()
{
  // construct a bank with 1 elevator and 2 floors
  elevators = new ElevatorBank(1, 2);
  elevatorInterface = elevators->getElevatorInterface();
  riderInterface = elevators->getRiderInterface();

  elevatorInterface->SetCallBack(this);
  riderInterface->SetCallBack(this);
  
  riderWakeup = new Semaphore("rider", 0);
  controllerWakeup = new Semaphore("controller", 0);
}

//----------------------------------------------------------------------
// ElevatorInspector::~ElevatorInspector
//	Deallocate data structures for testing the elevator simulation.
//----------------------------------------------------------------------

ElevatorInspector::~ElevatorInspector()
{ 
  delete elevators;
  delete riderWakeup;
  delete controllerWakeup;
}
        
//----------------------------------------------------------------------
// ElevatorInspector::RiderTest
//	Behave like a rider, going from floor 0 to floor 1.
//----------------------------------------------------------------------

void
ElevatorInspector::RiderTest ()
{
  int floor, elevator;
  ElevatorEvent event;
    
  riderInterface->PressButton(0, Up);
  event = WaitForNextRiderEvent(&floor, &elevator);
  ASSERT(event == DoorsOpened && floor == 0);
  ASSERT(riderInterface->getDirection(elevator) == Up);
  // should make sure that there aren't too many on the elevator
  // already
  ASSERT(riderInterface->EnterElevator(floor, elevator));
  riderInterface->PressFloor(1, elevator);
  event = WaitForNextRiderEvent(&floor, &elevator);
  ASSERT(event == DoorsOpened && floor == 1);
  ASSERT(riderInterface->ExitElevator(floor, elevator));
}

//----------------------------------------------------------------------
// ElevatorInspector::ControllerTest
//	Behave like an elevator controller, for testing the
//	elevator simulation.
//----------------------------------------------------------------------

void
ElevatorInspector::ControllerTest () {
  int floor, elevator;
  ElevatorEvent event;

  // assume elevator starts on ground floor    
  ASSERT(elevatorInterface->WhereIsElevator(0) == 0);	
  event = WaitForNextControllerEvent(&floor, &elevator);
  ASSERT(event == UpButtonPressed && floor == 0);
    
  // indicate to rider that elevator is going up
  elevatorInterface->MarkDirection(0, Up);	
  elevatorInterface->OpenDoors(0);
    
  // should wait a fixed amount of time for the rider to get
  // on, in case they took the stairs
  event = WaitForNextControllerEvent(&floor, &elevator);
  ASSERT(event == FloorButtonPressed && floor == 1);
  elevatorInterface->CloseDoors(0);
  elevatorInterface->MoveTo(floor, 0);
  event = WaitForNextControllerEvent(&floor, &elevator);
  ASSERT(event == ElevatorArrived && floor == 1 && elevator == 0);
  elevatorInterface->OpenDoors(elevator);
}

//----------------------------------------------------------------------
// ElevatorInspector::CallBack
//	Handle an event; since we don't know whether it is a
//	rider event or a controller event, just wake them both up.
//	If there's no event pending, they'll just go right back
//	to sleep.
//----------------------------------------------------------------------

void
ElevatorInspector::CallBack() {
  controllerWakeup->V();
  riderWakeup->V();
}
