/* syscalls.h 
 * 	Nachos system call interface.  These are Nachos kernel operations
 * 	that can be invoked from user programs, by trapping to the kernel
 *	via the "syscall" instruction.
 *
 *	This file is included by user programs and by the Nachos kernel. 
 *
 * Copyright (c) 1992-1993 The Regents of the University of California.
 * All rights reserved.  See copyright.h for copyright notice and limitation 
 * of liability and disclaimer of warranty provisions.
 */

#ifndef SYSCALL_H
#define SYSCALL_H

#include "copyright.h"

/* system call codes -- used by the stubs to tell the kernel which system call
 * is being asked for
 */
#define SC_agireq	-1
#define SC_halt		0
#define SC_exit		1
#define SC_exec		2
#define SC_join		3
#define SC_create	4
#define SC_open		5
#define SC_read		6
#define SC_write	7
#define SC_close	8
#define SC_mmap		9
#define SC_openport	10
#define SC_recv		12
#define SC_send		13
#define SC_closeport	14

/* extra credit?
#define SC_threadfork	15
#define SC_threadyield	16
*/

#ifndef IN_ASM

/* A unique identifier for an executing user program (address space) */
typedef int SpaceId;	
 
/* A unique identifier for an open Nachos file. */
typedef int OpenFileId;	

/* when an address space starts up, it has two open files, representing 
 * keyboard input and display output (in UNIX terms, stdin and stdout).
 * Read and Write can be used directly on these, without first opening
 * the console device.
 */
#define ConsoleInputId	0  
#define ConsoleOutputId	1

/* A connection endpoint is a remote machine combined with a port on that
 * machine. At the user program level, all network data is exchanged between
 * connection endpoints.
 */

typedef struct {
  int	machine;
  int	port;
} CEP;

#ifndef KERNEL
 
/* The system call interface.  These are the operations the Nachos
 * kernel needs to support, to be able to run user programs.
 *
 * Each of these is invoked by a user program by simply calling the 
 * procedure; an assembly language stub stuffs the system call code
 * into a register, and traps to the kernel.  The kernel procedures
 * are then invoked in the Nachos kernel, after appropriate error checking, 
 * from the system call entry point in exception.cc.
 */


/* For use by the auto-grader. Auto-grader test programs use this interface
 * to communicate with the kernel portion of the auto-grader. Never call this
 * in your programs.
 */
int agireq(int arg1, int arg2, int arg3, int arg4);


/* Stop Nachos, and print out performance stats */
void halt();		
 

/* Address space control operations: Exit, Exec, and Join */

/* This user program is done (status = 0 means exited normally). */
void exit(int status);	

/* Run the executable, stored in the Nachos file "name", and return the 
 * address space identifier, passing argc arguments in array argv
 */
SpaceId exec(char *name, int argc, char** argv);
 
/* Only return once the the user program "id" has finished.  
 * Return the exit status. Only the parent of the program can this.
 * Unlike Thread::Join, this is not necessarily ever called.
 */
int Join(SpaceId id); 	
 

/* File system operations: Create, Open, Read, Write, Close
 * These functions are patterned after UNIX -- files represent
 * both files *and* hardware I/O devices.
 *
 * If this assignment is done before doing the file system assignment,
 * note that the Nachos file system has a stub implementation, which
 * will work for the purposes of testing out these routines.
 */
 
/* Create a Nachos file, with "name" */
void create(char *name);

/* Open the Nachos file "name", and return an "OpenFileId" that can 
 * be used to read and write to the file.
 */
OpenFileId open(char *name);

/* Write "size" bytes from "buffer" to the open file. */
void write(OpenFileId id, char *buffer, int size);

/* Read "size" bytes from the open file into "buffer".  
 * Return the number of bytes actually read -- if the open file isn't
 * long enough, or if it is an I/O device, and there aren't enough 
 * characters to read, return whatever is available. For I/O devices,
 * you should NOT wait until bytes are available. This will make the chat
 * client in project 4 easier.
 */
int read(OpenFileId id, char *buffer, int size);

/* Close the file, we're done reading and writing to it. */
void close(OpenFileId id);

/* Map an open file to the specified address. Returns the size of the file. */
int mmap(OpenFileId id, char* address);

/* Allocate the port (a.k.a. mailbox) for the current process. A process that
   owns a port can send and receive data on that port.
   Returns 0 if successful, or -1 if the port number is invalid or already in
   use.
 */
int openport(int port);

/* Receive up to size bytes on this port, from ONE remote connection endpoint.
 * The remote connection endpoint from which the data is received is returned
 * in the remotemachine and remoteport pointers.
 * Returns the number of bytes transferred to the buffer if successful. If no
 * data is available on the port, this should return 0 immediately.
 */

int recv(int port, char* buffer, int size, CEP* remote);

/* Send size bytes from the port to the remote connection endpoint.
 */

void send(int port, char* buffer, int size, CEP* remote);
 
/* Release the port. The process gives up the ability to send and receive data
 * on the port.
 */

void closeport(int port);


// EXTRA CREDIT WILL NO LONGER BE GIVEN FOR IMPLEMENTING THESE

/* User-level thread operations: Fork and Yield.  To allow multiple
 * threads to run within a user program. 
 *
 * Could define other operations, such as LockAcquire, LockRelease, etc.
 */

/* Fork a thread to run a procedure ("func") in the *same* address space 
 * as the current thread.
 */
void threadfork(void (*func)(void* arg));

/* Yield the CPU to another runnable thread, whether in this address space 
 * or not. 
 */
void threadyield();		

#endif /* KERNEL */

#endif /* IN_ASM */

#endif /* SYSCALL_H */
